import datetime
import inspect
import os
import os.path
import sys

import test
from squish import snooze

squish_trace_ignored_file_names = []
squish_trace_command_delay_in_milliseconds = 0

squish_trace_function_exit = True

squish_trace_take_screenshots = False
squish_trace_screenshot_external_command = None
squish_trace_screenshot_folder = os.path.expanduser("~/Desktop/squish_trace_screenshots")
squish_trace_screenhot_file_name_prefix = "squish_trace-%s" % datetime.datetime.strftime(datetime.datetime.now(), "%Y%m%d-%H%M%S")
squish_trace_screenshot_count = 0



def squish_trace_log(msg, detail=None):
    if detail is None:
        test.warning(msg)
    else:
        test.warning(msg, detail)

def squish_trace_enable(command_delay_in_milliseconds=500, take_screenshots=False, trace_function_exit=True):
    squish_trace_log("TRACE: Command tracing enabled")
    sys.setprofile(squish_trace_function)
    squish_trace_set_command_delay_in_milliseconds(command_delay_in_milliseconds)
    squish_trace_set_take_screenshots(take_screenshots)
    global squish_trace_function_exit
    squish_trace_function_exit = trace_function_exit

def squish_trace_disable():
    squish_trace_log("TRACE: Command tracing disabled")
    sys.setprofile(None)
    squish_trace_set_take_screenshots(False)



def squish_trace_set_screenshot_external_command(screenshot_external_command):
    global squish_trace_screenshot_external_command
    squish_trace_screenshot_external_command = screenshot_external_command

def squish_trace_get_screenshot_external_command():
    global squish_trace_screenshot_external_command
    return squish_trace_screenshot_external_command

def squish_trace_set_screenshot_folder(screenshot_folder):
    global squish_trace_screenshot_folder
    squish_trace_screenshot_folder = screenshot_folder

def squish_trace_set_command_delay_in_milliseconds(command_delay_in_milliseconds):
    global squish_trace_command_delay_in_milliseconds
    squish_trace_command_delay_in_milliseconds = command_delay_in_milliseconds

def squish_trace_get_command_delay_in_milliseconds():
    global squish_trace_command_delay_in_milliseconds
    return squish_trace_command_delay_in_milliseconds

def squish_trace_set_take_screenshots(take_screenshots):
    global squish_trace_take_screenshots
    squish_trace_take_screenshots = take_screenshots



def squish_trace_set_ignored_file_names(file_names):
    global ignored_file_names
    ignored_file_names = file_names

def squish_trace_get_ignored_file_names():
    global squish_trace_ignored_file_names
    return squish_trace_ignored_file_names



def squish_trace_add_ignored_file_name(file_name):
    global squish_trace_ignored_file_names
    squish_trace_ignored_file_names.append(file_name)

def squish_trace_remove_ignored_file_name(file_name):
    global squish_trace_ignored_file_names
    squish_trace_ignored_file_names.remove(file_name)



def squish_trace_get_next_screenshot_info():
    global squish_trace_screenhot_file_name_prefix
    screenshot_file_name = squish_trace_screenhot_file_name_prefix
    global squish_trace_screenshot_count
    squish_trace_screenshot_count += 1
    screenshot_file_name += "-%04d.png" % squish_trace_screenshot_count
    global squish_trace_screenshot_folder
    return squish_trace_screenshot_folder, squish_trace_screenshot_folder + "/" + screenshot_file_name



def squish_trace_function(frame, event, arg):
    msg = "TRACE: >>    %s\n    args: %s\n    varargs: %s\n    keywords: %s\n    locals: %s\n\n"
    if event in ["return", "c_return"]:
        global squish_trace_function_exit
        if squish_trace_function_exit:
            msg = "TRACE: <<    %s\n    args: %s\n    varargs: %s\n    keywords: %s\n    locals: %s\n\n"
        else:
            return

    file_name = frame.f_code.co_filename
    if "python/" in file_name or \
        "python\\" in file_name or \
        "squish_trace.py" in file_name:
        return
    global squish_trace_ignored_file_names
    for n in squish_trace_ignored_file_names:
        if n in file_name:
            return

    lines, line_offset = inspect.getsourcelines(frame.f_code)
    line = lines[frame.f_lineno - line_offset][0:-1].strip()
    arg_names, vararg_names, keyword_names, args_and_locals = inspect.getargvalues(frame)
    args = {}
    varargs = {}
    keywords = {}
    locals = {}
    for n in args_and_locals.keys():
        v = args_and_locals[n]
        if n in arg_names:
            args[n] = v
        elif vararg_names is not None and n in vararg_names:
            varargs[n] = v
        elif keyword_names is not None and n in keyword_names:
            keywords[n] = v
        else:
            locals[n] = v

    # Since the details are shown in the control bar too, omit this for now:
    #s_args = squish_trace_dict_to_str(args)
    #s_varargs = squish_trace_dict_to_str(varargs)
    #s_keywords = squish_trace_dict_to_str(keywords)
    #s_locals = squish_trace_dict_to_str(locals)

    msg = msg % (line, args, varargs, keywords, locals)

    # Since the details are shown in the control bar too, omit this for now:
    #detail = "args:\n%svarargs:\n%skeywords:\n%slocals:\n%s\n" % (s_args, s_varargs, s_keywords, s_locals)
    detail = ""

    #msg += "   (%s)" % event
    #detail = "%s:%s" % (frame.f_code.co_filename, frame.f_lineno)

    global squish_trace_take_screenshots
    if squish_trace_must_take_screenshot():
        screenshot_folder, screenshot_file_name = squish_trace_get_next_screenshot_info()
        detail += "Screenshot: %s\n\n" % screenshot_file_name
    squish_trace_log(msg, detail)

    # Only delay if:
    # - Screenshots are disabled, or
    # - A screenshot must be taken, but via saveDesktopScreenshot()
    #   (as opposed to an external tool) because
    #   saveDesktopScreenshot() takes about 3 seconds to complete...
    delay = squish_trace_get_command_delay_in_milliseconds()
    if squish_trace_must_take_screenshot() and \
        squish_trace_get_screenshot_external_command() is None:
        delay -= 3000

    snooze_before_screenshot = True
    if delay > 0:
        snooze_before_screenshot = False
        snooze(delay / 1000.0)

    if squish_trace_must_take_screenshot():
        if not os.path.exists(screenshot_folder + "/"):
            os.mkdir(screenshot_folder)

        # Snooze here to give control bar time to update, not
        # very reliable in achieving that anyway:
        if snooze_before_screenshot:
            snooze(0.2)
        squish_trace_take_screenshot(screenshot_file_name)

def squish_trace_dict_to_str(d, line_prefix="    "):
    keys = d.keys()
    keys.sort()
    s = ""
    for k in keys:
        s += "%s%s: %s\n" % (line_prefix, k, d[k])
    return s

def squish_trace_must_take_screenshot():
    global squish_trace_take_screenshots
    res = squish_trace_take_screenshots and len(applicationContextList()) > 0
    return res

def squish_trace_take_screenshot(screenshot_file_name):
    ext_cmd = squish_trace_get_screenshot_external_command()
    if ext_cmd is None:
        saveDesktopScreenshot(screenshot_file_name)
    else:
        cmd = ext_cmd % screenshot_file_name
        os.system(cmd)
